#!/usr/bin/perl
# 
# Generate NICER background PI spectrum based on "3C50_2020a" library
#
my $taskname = "nibackgen3C50.pl";
my $taskvers = "0.6";
# v0.0 - initial version
# v0.1 - filters out HBG outliers
# v0.2 - enable list of FPMs to exclude
# v0.3 - incorporates some HEASoft 6.25-to-6.26.1 changes of other NICER  tasks
use HEACORE::HEAINIT;
# v0.4 - bug fix for many-GTI case, reduce number of "tmp" files created
# v0.5 - bug fix in mathpha_str subroutine
# v0.6 - changes to support different epoch libraries
# v0.7 - added s0 cut

# ===================================
# Execute main subroutine, with error trapping
$status = 0;
my @tmpfiles = ();
my ($tmpfile, $atstatus, $cleanup);
$cleanup = 1;
eval {
    $status = headas_main(\&nibackgen3C50);
};
$atstatus = $@;

# ===================================
# Check for errors and report them to the console
if ($atstatus) {
    if ($status == 0) { $status = -1; }
    warn $atstatus;
}
exit $status;
exit 0;
# <========= Main task ends here

# ===================================
# Main subroutine

sub nibackgen3C50 {
    # Makes all environment variables available
    use Env;
    use Cwd;

    # The HEAUTILS module provides access to HDpar_stamp()
    # set_toolname(), set_toolversion(), and headas_clobberfile()
    use HEACORE::HEAUTILS;
    use HEACORE::PIL;
    # include the file specification functions
    use Astro::FITS::CFITSIO qw( :shortnames :constants );
    use POSIX qw(fmod);

    # Use the standard HEAdas methods for registering the toolname and version number to be
    # used in error reporting and in the record of parameter values written by HDpar_stamp
    set_toolname($taskname);
    set_toolversion($taskvers);

    # Prevent subsidiary tasks from pausing for input 
    $ENV{"HEADASNOQUERY"} = 1;
    $ENV{"HEADASPROMPT"} = "/dev/null";

    eval {
	$status = &nibackgen3C50_work();
    };

    # Special effort to remove scratch files
    if ($cleanup) {
	foreach $scratchfile (@tmpfiles) {
	    unlink "$scratchfile" if (-f $scratchfile);

	}
    }

    if ($@) {
	if ($status == 0) { $status = -1; }
	warn $@;
	return $status;
    }

    return $status;
}

sub nibackgen3C50_work {
    # Makes all environment variables available
    use Env;
    use Cwd;

    # The HEAUTILS module provides access to HDpar_stamp()
    # set_toolname(), set_toolversion(), and headas_clobberfile()
    use HEACORE::HEAUTILS;
    use HEACORE::PIL;
    # include the file specification functions
    use Astro::FITS::CFITSIO qw( :shortnames :constants );
    # User defined module which contains the Perl-CFITSIO interface functions 
    use SimpleFITS;
    use POSIX qw(fmod);

    # --------------- parameters
    my $chatter;
    $status = PILGetInt("chatter",$chatter);
    my $verbose = ($chatter >= 5)?(1):(0);
    # Ordered parameters, usually ask parameters which must come first
    my @parmlist = ("rootdir");
    my %parms = (
		 'rootdir' => \&PILGetString,
		 'obsid' => \&PILGetString,
	         'bkgidxdir' => \&PILGetString,
	         'bkglibdir' => \&PILGetString,
		 'gainepoch' => \&PILGetString,
		 'calevtdir' => \&PILGetString,
		 'totspec' => \&PILGetString,
		 'bkgspec' => \&PILGetString,
		 'ufafile' => \&PILGetString,
		 'clfile' => \&PILGetString,
		 'fpmofflist' => \&PILGetString,
		 'dtmin' => \&PILGetReal,
		 'dtmax' => \&PILGetReal,
	         'hbgcut' => \&PILGetReal,
		 's0cut' => \&PILGetReal,
		 'minweight' => \&PILGetReal,
		 'cleanup' => \&PILGetBool,
	         'clobber' => \&PILGetBool,
		 'history' => \&PILGetBool,
		);
    # Find parameter values (listed above)
    my ($parm, $func, $val);
    # ... first read ordered parameters, then anything else
    foreach $parm ( @parmlist, keys(%parms) ) {
      my $func = $parms{$parm};
      next if (ref($func) ne "CODE");  # Skip if we already did this parm
      undef($val);
      $status = &$func("$parm", $val);
      die "ERROR: could not retrieve parameter '$parm'" if ($status);
      $parms{$parm} = $val;
      print "$parm=$val\n" if ($verbose);
    }

    my $clobstr = ($parms{clobber}) ? "YES" : "NO";
    my $histstr = ($parms{history}) ? "YES" : "NO";
    # Global variable
    $cleanup = $parms{cleanup};
    my $cleanstr = ($cleanup) ? "YES" : "NO";

    print "$taskname $taskvers\n" if ($chatter >= 1);
    print "--------------------------------------------------------\n" if ($chatter >= 1);
    foreach my $k (sort keys(%parms)) {
      print "      $k = ".$parms{$k}."\n" if ($chatter >= 2);
    }

    my $rootdir = $parms{rootdir};
    my $obsid = $parms{obsid};

    if ( ($parms{rootdir} !~ m/^NONE$/i) && ($params{obsid} !~ m/^NONE$/i) ) {
        $indir = $rootdir . "/" . $obsid;
	die "ERROR: input directory $indir does not exist" if (! -d $indir);
    }

    my $calevtdir = $parms{calevtdir};
    if ($parms{calevtdir} !~ m/^NONE$/i) {
	$calevtdir =~ s/\$OBSID/$obsid/g; $calevtdir =~ s/\$INDIR/$indir/g;
	die "ERROR: cl directory $calevtdir does not exist" if (! -d $calevtdir);
    }

    # Find the input event files
    my $ufafile = $parms{ufafile};
    my $clfile = $parms{clfile};
    $ufafile =~ s/\$CALEVTDIR/$calevtdir/g; $ufafile =~ s/\$OBSID/$obsid/g; $ufafile =~ s/\$INDIR/$indir/g;
    $ufafile .= ".gz" if ( ! -f $ufafile );
    die "ERROR: uf file $ufafile does not exist" if (! -f $ufafile);
    $clfile  =~ s/\$CALEVTDIR/$calevtdir/g; $clfile  =~ s/\$OBSID/$obsid/g; $clfile  =~ s/\$INDIR/$indir/g;
    $clfile .= ".gz" if ( ! -f $clfile );
    die "ERROR: cl file $clfile does not exist" if (! -f $clfile);

    # Extract the GAIN CALDB file used for later comparison with library version
    $cldata = SimpleFITS->open("$clfile", type=>"table", access=>"read");
    $cldata->move("EVENTS");
    $cldata->readkey("GCALFILE",$gcalfile);
    $cldata->close();

    my $gcalidx = index($gcalfile, "fits");
    if ($gcalidx != -1) {
	$gcompare = substr $gcalfile, 0, $gcalidx+4;
    } else {
	$gcompare = $gcalfile;
    }
    #print "GCALFILE: $gcalfile GALIDX: $gcalidx GCOMP: $gcompare\n";
    
    # Background index file directory
    my $bkgidxdir = $parms{bkgidxdir};
    die "ERROR: bkg index file directory $bkgidxdir does not exist" if (! -d $bkgidxdir);
    my $bkgidxfilename = "nibackgen3C50_info.fits";
    $bkgidxfile = $bkgidxdir . "/" . $bkgidxfilename;
    die "ERROR: bkg index file $bkgidxfile does not exist" if (! -f $bkgidxfile);
    $bkgidxdata =  SimpleFITS->open("$bkgidxfile", type=>"table", access=>"read");
    die "ERROR: Failed to open background index file" if (! $bkgidxdata );
    $bkgidxdata->move("BGINDEX");
    @years = $bkgidxdata->readcol("YEAR",{type=>TSTRING});
    
    # Gain epoch and gain epoch index
    my $gainepoch = $parms{gainepoch};
    # Check that gain epoch is one of the values in the index file
    die "ERROR: NOT an allowed gain epoch" if (not ( ( grep { $_ eq $gainepoch} @years ) ) );
    my ($yridx) = grep { $years[$_] eq $gainepoch } (0 .. @years-1);
    
    # Check that file used for GAIN application is appropriate
    @gainfiles = $bkgidxdata->readcol("GAINFILE",{type=>TSTRING});
    my $gainfile = $gainfiles[$yridx];
    if ($gainfile ne $gcompare) {
	print "WARNING: GCALFILE keyword in $clfile, $gcalfile, does not match gain file used to construct background library, $gainfile. This warning may be disregarded in cases where the gainepoch parameter is set to 2020 and the data is processed using CALDB gain file nixtiflightpi20170601v006.fits or later.\n";
    }
    
    my $totspec = $parms{totspec}.".pi";
    my $bkgspec = $parms{bkgspec}.".pi";
    my $bkgspec_ngt = $parms{bkgspec}."_ngt.pi";
    my $bkgspec_day = $parms{bkgspec}."_day.pi";
    
    my $fpms_off = ();
    if ($parms{fpmofflist} !~ m/^none$/i) {
	my @fpmlist = parse_fpm_list($parms{fpmofflist});
	print "  FPM List: $parms{fpmofflist}\n" if ($chatter >= 2);
	print "     --> ".join(',',@fpmlist)."\n" if ($chatter >= 4);
	for my $fpm (@fpmlist) {
	    push @fpms_off, $fpm;
	}
    }
    
    # Background library (sub)directory
    @subdirs = $bkgidxdata->readcol("SUBDIR",{type=>TSTRING});
    my $subdir = $subdirs[$yridx];
    my $bkglibdir = $parms{bkglibdir} . "/" . $subdir;
    die "ERROR: bkg directory $bkglibdir does not exist" if (! -d $bkglibdir);
    # Background library specs
    @ibgmins = $bkgidxdata->readcol("IBGMIN",{type=>TDOUBLE});
    my $ibgmin = $ibgmins[$yridx];
    @hrejmaxs = $bkgidxdata->readcol("HREJMAX",{type=>TDOUBLE});
    my $hrejmax = $hrejmaxs[$yridx];
    @trmpexprs = $bkgidxdata->readcol("TRMPEXPR",{type=>TSTRING});
    my $trmpexpr = $trmpexprs[$yridx];
    @hrejexprs = $bkgidxdata->readcol("HREJEXPR",{type=>TSTRING});
    my $hrejexpr = $hrejexprs[$yridx];
    @ibgexprs = $bkgidxdata->readcol("IBGEXPR",{type=>TSTRING});
    my $ibgexpr = $ibgexprs[$yridx];
    @hbgexprs = $bkgidxdata->readcol("HBGEXPR",{type=>TSTRING});
    my $hbgexpr = $hbgexprs[$yridx];
    @nzexprs = $bkgidxdata->readcol("NZEXPR",{type=>TSTRING});
    my $nzexpr = $nzexprs[$yridx];
    @s0exprs = $bkgidxdata->readcol("S0EXPR",{type=>TSTRING});
    my $s0expr = $s0exprs[$yridx];
    print "IBGMIN: $ibgmin HREJMAX: $hrejmax\n" if ($chatter >= 3);
    print "TRUMPET EXPR: $trmpexpr\n" if ($chatter >= 3);
    print "HREJ EXPR: $hrejexpr\n" if ($chatter >= 3);
    print "IBG EXPR: $ibgexpr\n" if ($chatter >= 3);
    print "HBG EXPR: $hbgexpr\n" if ($chatter >= 3);
    print "NZ EXPR: $nzexpr\n" if ($chatter >= 3);
    print "S0 EXPR: $s0expr\n" if ($chatter >= 3);
    
    # Background library data - Night
    @ngtexts = $bkgidxdata->readcol("NGTEXT",{type=>TSTRING});
    my $ngtext = $ngtexts[$yridx];
    print "Moving to extension: $ngtext\n" if ($chatter >= 3);
    $bkgidxdata->move($ngtext);
    $bkgidxdata->readkey("NAXIS2",$nibg);
    @bg_ngt_names = $bkgidxdata->readcol("SPECNAME",{type=>TSTRING});
    @bg_ngt_ids = $bkgidxdata->readcol("BKGID",{type=>TSTRING});
    @bg_ngt_ibgs = $bkgidxdata->readcol("IBGAVG",{type=>TDOUBLE});
    @ibglo = $bkgidxdata->readcol("IBGLO",{type=>TDOUBLE});
    @ibghi = $bkgidxdata->readcol("IBGHI",{type=>TDOUBLE});
    @hrejlo = $bkgidxdata->readcol("HREJLO",{type=>TDOUBLE});
    @hrejhi = $bkgidxdata->readcol("HREJHI",{type=>TDOUBLE});
    @hbglibrate = $bkgidxdata->readcol("HBGRATE",{type=>TDOUBLE});
    @s0librate_ngt = $bkgidxdata->readcol("S0NGTRATE",{type=>TDOUBLE});
    my $ibgmax = $ibghi[nibg-1];
    print "NIBG: $nibg IBGMAX: $ibgmax\n" if ($chatter >= 3);
    my @bg_ngt_expos = (0) x ($nibg);
    for (my $ii=0; $ii<$nibg; $ii+=1) {
	$bg_ngt_spec = SimpleFITS->open("$bkglibdir" . "/" . "$bg_ngt_names[$ii]", type=>"table", access=>"read");
	$bg_ngt_spec->move("SPECTRUM");
	$bg_ngt_spec->readkey("EXPOSURE",$bg_ngt_expo);
	$bg_ngt_expos[$ii] = $bg_ngt_expo;
        print "SPEC: $bg_ngt_names[$ii] ID: $bg_ngt_ids[$ii] IBGAVG: $bg_ngt_ibgs[$ii] IBGLO: $ibglo[$ii] IBGHI: $ibghi[$ii] HREJLO : $hrejlo[$ii] HREJHI: $hrejhi[$ii] HBGRATE: $hbglibrate[$ii] S0RATE: $s0librate_ngt[$ii] EXPO: $bg_ngt_expos[$ii]\n" if ($chatter >= 3);
    }
    
    # Background library data - Day
    $bkgidxdata->move("BGINDEX");
    @dayexts = $bkgidxdata->readcol("DAYEXT",{type=>TSTRING});
    my $dayext = $dayexts[$yridx];
    print "Moving to extension: $dayext\n" if ($chatter >= 3);
    $bkgidxdata->move($dayext);
    $bkgidxdata->readkey("NAXIS2",$nnz);
    my @bg_day_names = $bkgidxdata->readcol("SPECNAME",{type=>TSTRING});
    my @bg_day_ids = $bkgidxdata->readcol("BKGID",{type=>TSTRING});
    my @bg_day_nzs = $bkgidxdata->readcol("NZAVG",{type=>TDOUBLE});
    my @nzlo = $bkgidxdata->readcol("NZLO",{type=>TDOUBLE});
    my @nzhi = $bkgidxdata->readcol("NZHI",{type=>TDOUBLE});
    my $nzmax = $nzhi[nnz-2];
    @s0librate_day = $bkgidxdata->readcol("S0DAYRATE",{type=>TDOUBLE});
    print "NNZ: $nnz NZMAX: $nzmax\n" if ($chatter >= 3);
    my @bg_day_expos = (0) x ($nnz);
    for (my $ii=0; $ii<$nnz; $ii+=1) {
	$bg_day_spec = SimpleFITS->open("$bkglibdir" . "/" . "$bg_day_names[$ii]", type=>"table", access=>"read");
	$bg_day_spec->move("SPECTRUM");
	$bg_day_spec->readkey("EXPOSURE",$bg_day_expo);
	$bg_day_expos[$ii] = $bg_day_expo;
	print "SPEC: $bg_day_names[$ii], ID: $bg_day_ids[$ii], NZAVG: $bg_day_nzs[$ii], NZLO: $nzlo[$ii], NZHI: $nzhi[$ii], S0RATE: $s0librate_day[$ii] EXPO: $bg_day_expos[$ii]\n" if ($chatter >= 3);
    }
    $bkgidxdata->close();

    $cmd = "ftcopy infile='$ufafile' ".
	"outfile='tmp_ufafile.evt' ".
	"copyall='yes' ".
	"clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    $ufagtistatus = CheckGTIName("tmp_ufafile.evt");
    die "ERROR: no GTI or STDGTI extension found in input ufa file" if ($ufagtistatus==-1);
    if ($cleanup) { push @tmpfiles, "tmp_ufafile.evt" };
    $cmd = "ftcopy infile='$clfile' ".
	"outfile='tmp_clfile.evt' ".
	"copyall='yes' ".
	"clobber='yes' "."history='$histstr' "."chatter='$chatter' ";  
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);  
    $clgtistatus = CheckGTIName("tmp_clfile.evt");
    die "ERROR: no GTI or STDGTI extension found in input cl file" if ($clgtistatus==-1);
    if ($cleanup) { push @tmpfiles, "tmp_clfile.evt" };

    # Filter bad detectors
    if ($parms{fpmofflist} =~ m/^none$/i) {
	$cmd = "ftcopy infile='tmp_ufafile.evt' ".
	    "outfile='tmp_ufafile_nobaddet.evt' ".
	    "copyall='yes' ".
	    "clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    } else {
	my $detcnt = 0;
	my $detexpr = "";
	my $first = 1;
	foreach $fpm (@fpms_off) {
	    if ($first==1) {
		$expr_to_add = "(DET_ID!=".$fpm.")";
		$first = 0;
	    } else {
		$expr_to_add = ".and.(DET_ID!=".$fpm.")";
	    }
	    $detexpr = $detexpr.$expr_to_add 
	}		
	$cmd = "ftcopy infile='tmp_ufafile.evt"."[EVENTS][".$detexpr."]' ".
	    "outfile='tmp_ufafile_nobaddet.evt' ".
	    "copyall='yes' ".
	    "clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    }
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: ftcopy failed to filter bad detectors" if ($?);
    die "ERROR: file with bad detectors filtered not produced" if (! -f "tmp_ufafile_nobaddet.evt");
    if ($cleanup) { push @tmpfiles, "tmp_ufafile_nobaddet.evt" };
    # Scaling factor, since library is scaled to 52 detectors
    my $nbad = scalar @fpms_off;
    my $livedet = 52;
    my $baddetrat = ($livedet-$nbad)/$livedet;
    print "NUMBER LIVE DETECTORS: $livedet NUMBER BAD DETECTORS: $nbad RATIO GOOD_TO_TOTAL: $baddetrat\n" if ($chatter >= 3);

    # Extract slow chain ufa file
    $cmd = "niextract-events ".
      "filename='tmp_ufafile_nobaddet.evt[PI=-10:1800,EVENT_FLAGS=bxxx1x000]' ".
      "eventsout='tmp_ufafile_slowchain.evt' ".
      "regionfile='NONE' timefile='NONE' xcolf='DETX' ycolf='DETY' tcol='TIME' ".
      "events='EVENTS' gti='GTI' gtimerge='OR' "."colexpr='NONE' ".
      "minselectmb='2000' "."copyall='no' "."cleanup='$cleanstr' ".
      "clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: task niextract-events failed to create slow chain ufa file" if ($?);
    die "ERROR: slow chain ufa file not produced" if (! -f "tmp_ufafile_slowchain.evt");
    if ($cleanup) { push @tmpfiles, "tmp_ufafile_slowchain.evt" };

    # Filter slow chain ufa file to create trumpet-selected ufa file
    $cmd = "ftcopy ".
	"infile='tmp_ufafile_slowchain.evt[EVENTS][".$trmpexpr."]' ".
	"outfile='tmp_ufafile_tsel.evt' ".
	"copyall='yes' ".
	"clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: ftcopy failed to create trumpet-selected ufa file" if ($?);
    die "ERROR: trumpet-selected ufa file not produced" if (! -f "tmp_ufafile_tsel.evt");
    if ($cleanup) { push @tmpfiles, "tmp_ufafile_tsel.evt" };

    # Filter trumpet-selected ufa file over IBG energies
    $cmd = "ftcopy ".
      "infile='tmp_ufafile_tsel.evt[EVENTS][".$ibgexpr."]' ".
      "outfile='tmp_ufafile_tsel_ibg.evt' ".
      "copyall='yes' ".
      "clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: ftcopy failed to create trumpet-selected IBG ufa file" if ($?);
    die "ERROR: trumpet-selected IBG ufa file not produced" if (! -f "tmp_ufafile_tsel_ibg.evt");
    if ($cleanup) { push @tmpfiles, "tmp_ufafile_tsel_ibg.evt" };

    # Filter trumpet-selected ufa file over HBG energies
    $cmd = "ftcopy ".
      "infile='tmp_ufafile_tsel.evt[EVENTS][".$hbgexpr."]' ".
      "outfile='tmp_ufafile_tsel_hbg.evt' ".
      "copyall='yes' ".
      "clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: ftcopy failed to create trumpet-selected HBG ufa file" if ($?);
    die "ERROR: trumpet-selected HBG ufa file not produced" if (! -f "tmp_ufafile_tsel_hbg.evt");
    if ($cleanup) { push @tmpfiles, "tmp_ufafile_tsel_hbg.evt" };

    # Filter trumpet-selected ufa file over S0 energies
    $cmd = "ftcopy ".
      "infile='tmp_ufafile_tsel.evt[EVENTS][".$s0expr."]' ".
      "outfile='tmp_ufafile_tsel_s0.evt' ".
      "copyall='yes' ".
      "clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: ftcopy failed to create trumpet-selected S0 ufa file" if ($?);
    die "ERROR: trumpet-selected S0 ufa file not produced" if (! -f "tmp_ufafile_tsel_s0.evt");
    if ($cleanup) { push @tmpfiles, "tmp_ufafile_tsel_s0.evt" };

    # Filter slow chain ufa file to create hatchet-rejected ufa file
    $cmd = "ftcopy ".
      "infile='tmp_ufafile_slowchain.evt[EVENTS][".$hrejexpr."]' ".
      "outfile='tmp_ufafile_hrej.evt' ".
      "copyall='yes' ".
      "clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: ftcopy failed to create hatchet-rejected ufa file" if ($?);
    die "ERROR: hatchet-rejecte ufa file not produced" if (! -f "tmp_ufafile_hrej.evt");
    if ($cleanup) { push @tmpfiles, "tmp_ufafile_hrej.evt" };

    # Filter trumpet-selected ufa file over NZ energies to get slow-chain noise file
    $cmd = "ftcopy ".
      "infile='tmp_ufafile_tsel.evt[EVENTS][".$nzexpr."]' ".
      "outfile='tmp_ufafile_tsel_nz.evt' ".
      "copyall='yes' ".
      "clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: ftcopy failed to create slow chain noise ufa file" if ($?);
    die "ERROR: slow chain noise ufa file not produced" if (! -f "tmp_ufafile_tsel_nz.evt");
    if ($cleanup) { push @tmpfiles, "tmp_ufafile_tsel_nz.evt" };

    # Redefine GTI attached to clfile so that for all final dt, dtmin<dt<=dtmax
    $makegtierror = createStartStopGTI("tmp_clfile.evt",$parms{dtmin},$parms{dtmax},$histstr,$chatter);
    die "ERROR: GTI re-definition FAILED" if ($makegtierror==-1);
    
    # Apply GTI to ufa file (without bad detectors)
    $cmd = "nicerclean ".
      "infile='tmp_ufafile_nobaddet.evt' ".
      "outfile='tmp_clfile_nobaddet_newgti.evt' ".
      "gtifile='tmp_newgti.gti' ".
      "pirange='20:1500' ".
      "trumpetfilt='yes' ".
      "filtexpr='EVENT_FLAGS=bxxx1x000' ".
      "cleanup='$cleanstr' "."history='$histstr' "."chatter='$chatter' "."clobber='yes' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: nicerclean failed to create new cleaned event file" if ($?);    
    die "ERROR: new cleaned event file not produced" if (! -f "tmp_clfile_nobaddet_newgti.evt");
    if ($cleanup) { push @tmpfiles, "tmp_clfile_nobaddet_newgti.evt" };

    $subgtierror = SubstituteGTI("tmp_clfile_nobaddet_newgti.evt","GTI","tmp_newgti.gti",$histstr,$chatter);
    die "ERROR: GTI substitution FAILED" if ($subgtierror==-1);
    
    # Get the number of GTI and delta-times in the event file with redifined GTI
    $newclean = SimpleFITS->open("tmp_clfile_nobaddet_newgti.evt", type=>"table", access=>"read");
    $newclean->move("EVENTS");
    $newclean->readkey("NAXIS2",$numnewclean);
    $newclean->readkey("EXPOSURE",$expnewclean);
    die "ERROR: no events in new cleaned event file" if ( $numnewclean<=0  );
    die "ERROR: zero good exposure in new cleaned event file" if ( $expnewclean<=0  );
    $newclean->move("GTI");
    $newclean->readkey("NAXIS2",$ngti);
    die "ERROR: no GTI in new cleaned event file" if ( $ngti<=0  );
    @indtgti = $newclean->calculate("STOP-START",{type=>TDOUBLE});
    $newclean->close();

    # initialize weight arrays
    my @weights = (0) x ($nibg);
    my @exp_weights = (0) x ($nibg);
    my @supweights = (0) x ($nnz);
    my @supexp_weights = (0) x ($nnz);
    my $exp_bkg = 0.0;

    # temp text file for storing list of GTI with rates in the library range
    my $tmpgti = "tmp_gti.gti";
    # start with the full GTI and delete rows as appropriate
    $cmd = "ftcopy infile='tmp_newgti.gti' ".
	"outfile='$tmpgti' ".
	"copyall='yes' ".
	"clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: failed to copy redefined GTI file" if ($?);
    my $tmpgtilst = "tmp_gtilist.txt";
    open GTILIST, ">", $tmpgtilst;

    # some bookeeping:
    my $nkeep = 0;
    my $nreject = 0;
    my $tkeep = 0.0;
    my $treject = 0.0;
    my $sumibg = 0.0;
    my $sumhrej = 0.0;
    my $sumhbg = 0.0;
    my $sums0 = 0.0;
    # loop over rows in gti file
    for (my $ii=1; $ii<$ngti+1; $ii+=1) { 
	my $ibgrate = 0.0;
	my $hrejrate = 0.0;
	my $hbgrate = 0.0;
	my $s0rate = 0.0;
	my $nzrate = 0.0;
	my $id12 = "00";
	my $id33 = "00";
	my $keep_flag = 0;
	my $s0flag = 0;
	print "X1: $ii $keep_flag $s0flag\n" if ($verbose); 
        # extract this row in the gti 
        $cmd = "ftcopy infile='tmp_clfile_nobaddet_newgti.evt[GTI][\#row ==$ii]' ".
          "outfile='tmp_clfile_nobaddet_newgti_$ii.gti' ".
          "copyall='no' ".
          "clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
	print "$cmd\n" if ($chatter >= 3);
	system($cmd);
	die "ERROR: failed to extract GTI row $ii" if ($?);
	die "ERROR: ftcopy failed to produce GTI row $ii" if (! -f "tmp_clfile_nobaddet_newgti_$ii.gti");
        if ($cleanup) { push @tmpfiles, "tmp_clfile_nobaddet_newgti_"."$ii".".gti" };
        # Best match and scaling factor for bkg spectrum - night
        $ibgrate = RateGTI("tmp_ufafile_tsel_ibg.evt","tmp_clfile_nobaddet_newgti_$ii.gti",$indtgti[$ii-1],$baddetrat,$histstr,$chatter);
	if ($ibgrate < $ibgmin) {
	    $ibgrate = $ibgmin;
	}
        $hrejrate = RateGTI("tmp_ufafile_hrej.evt","tmp_clfile_nobaddet_newgti_$ii.gti",$indtgti[$ii-1],$baddetrat,$histstr,$chatter); 
        $hbgrate = RateGTI("tmp_ufafile_tsel_hbg.evt","tmp_clfile_nobaddet_newgti_$ii.gti",$indtgti[$ii-1],$baddetrat,$histstr,$chatter);
	$s0rate = RateGTI("tmp_ufafile_tsel_s0.evt","tmp_clfile_nobaddet_newgti_$ii.gti",$indtgti[$ii-1],$baddetrat,$histstr,$chatter);
	$id12 = NewFindIBGID(\@bg_ngt_ids,\@ibglo,\@ibghi,\@hrejlo,\@hrejhi,\$nibg,\$ibgmax,\$hrejmax,\$ibgrate,\$hrejrate);
	print "IBG: $ibgrate HREJ: $hrejrate ID12: $id12\n" if ($verbose);
	if ($id12 ne "00") {
	    my $bgngtindx = -1;
	    $bgngtindx++ until @bg_ngt_ids[$bgngtindx]==$id12 or $bgngtindx>$nibg;
	    $ibgbkg = $bg_ngt_ibgs[$bgngtindx];
	    # the rate corrected for number of MPU is used for matching, 
	    # but the uncorrected rate is used for scaling
	    $ibgrate_uncorr = $baddetrat * $ibgrate;
	    $hbgrate_uncorr = $baddetrat * $hbgrate;
	    $hbg_bkg = ($ibgrate_uncorr / $ibgbkg) * $hbglibrate[$bgngtindx];
	    $hbg_tot = abs($hbgrate_uncorr-$hbg_bkg);
	    $s0rate_uncorr = $baddetrat * $s0rate;
	    $s0_bkg = ($ibgrate_uncorr / $ibgbkg) * $s0librate_ngt[$bgngtindx];
	    $s0_tot = $s0rate_uncorr - $s0_bkg;
	    print "IDX: $bgngtindx IBGAVG: $ibgbkg IBG_UNCORR: $ibgrate_uncorr S0_UNCORR: $s0rate_uncorr HBG_UNCORR: $hbgrate_uncorr HBG_BKG: $hbg_bkg HBG_TOT: $hbg_tot s0_BKG: $s0_bkg S0_TOT: $s0_tot\n" if ($verbose);
	    # proceed only if residual HBG is not too high
	    print "X2: $keep_flag $s0flag\n" if ($verbose);
	    if ($hbg_tot < $parms{hbgcut}) {
		$s0flag = 1;
		# Best match bkg spectrum and scaling factor - day correction
		$nzrate = RateGTI("tmp_ufafile_tsel_nz.evt","tmp_clfile_nobaddet_newgti_$ii.gti",$indtgti[$ii-1],$baddetrat,$histstr,$chatter);
		$id33 = NewFindNZID(\@bg_day_ids,\@nzlo,\@nzhi,\$nnz,\$nzmax,\$nzrate);
		print "NZ: $nzrate ID33: $id33\n" if ($verbose);
		print "X3: $keep_flag $s0flag\n" if ($verbose);
		if ($id33 ne "00") {
		    my $bgdayindx = -1;
		    $bgdayindx++ until @bg_day_ids[$bgdayindx]==$id33 or $bgdayindx>$nnz;
		    $nzbkg = $bg_day_nzs[$bgdayindx];
		    $nzrate_uncorr = $baddetrat * $nzrate;
		    $s0_bkg += ($nzrate_uncorr / $nzbkg) * $s0librate_day[$bgdayindx];
		    $s0_tot -= ($nzrate_uncorr / $nzbkg) * $s0librate_day[$bgdayindx];
		    print "$ii IDX: $bgdayind S0_UNCORR: $s0rate_uncorr S0_BKG: $s0_bkg S0_TOT: $s0_tot\n" if ($verbose);
		    print "X4: $keep_flag $s0flag\n" if ($verbose); 
		    if (abs($s0_tot) < $parms{s0cut}) {
			$dsupweight = $indtgti[$ii-1] * $nzrate_uncorr / $nzbkg;
			$supweights[$bgdayindx] += $dsupweight;
			$supexp_weights[$bgdayindx] += $indtgti[$ii-1];
		    } else {
			$s0flag = 0;
		    }
		    print "X4: $keep_flag $s0flag\n" if ($verbose);
		} else {
		    print "No ISS day supplement needed for refactored GTI row $ii\n" if ($chatter >= 1);
		}
		if ($s0flag == 1) {
		    $keep_flag = 1;
		    print "X5: $keep_flag $s0flag\n" if ($verbose);
		    $dweight = $indtgti[$ii-1] * $ibgrate_uncorr / $ibgbkg;
		    $weights[$bgngtindx] += $dweight;
		    $exp_weights[$bgngtindx] += $indtgti[$ii-1];
		    print GTILIST "tmp_clfile_nobaddet_newgti_$ii.gti\n";
		} else {
		    print "Residual 0.2-0.3 keV rate exceeds user-defined cutoff; no background estimate contribution from refactored GTI row $ii\n" if ($chatter >= 1);
		}
	    } else {
		print "Residual 13-15 keV rate exceeds user-defined cutoff; no background estimate contribution from refactored GTI row $ii\n" if ($chatter >= 1);
	    }

	}
	if ($keep_flag == 1) {
            $nkeep += 1;
            $tkeep += $indtgti[$ii-1];
            $sumibg += $ibgrate*$indtgti[$ii-1];
            $sumhrej += $hrejrate*$indtgti[$ii-1];
	    $sumhbg += $hbgrate*$indtgti[$ii-1];
	    $sums0 += $s0rate*$indtgti[$ii-1];
        } else {
            $nreject += 1;
            $treject += $indtgti[$ii-1];
	    # remove this GTI from event file used to extract the total spectrum
            my $ndelete = $nkeep+1;
	    $cmd = "ftdelrow infile='$tmpgti"."[GTI]' ".
		"outfile='$tmpgti' ".
		"rows='$ndelete' ".
		"confirm='yes' ".
		"clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
	    print "$cmd\n" if ($chatter >= 3);
	    system($cmd); 
	    die "ERROR: failed to delete row from redefined GTI file" if ($?)
	}
	unlink "tmp_clfile_nobaddet_newgti_$ii.gti";
    } #end loop over GTI
    $exp_bkg += $tkeep;
    if ($nkeep > 0) {
        $ibgav =  $sumibg / $tkeep;
        $hrejav =  $sumhrej / $tkeep;
        $hbgav =  $sumhbg / $tkeep;
	$s0av =  $sums0 / $tkeep;
        print "Exposure time used: $tkeep, Exposure time discarded: $treject\n";
        print "Average IBG rate in used GTI: $ibgav\n";
        print "Average HREJ rate in used GTI: $hrejav\n";
        print "Average HBG rate in used GTI: $hbgav\n";
	print "Average S0 rate in used GTI: $s0av\n";
	close GTILIST;
        if ($cleanup) { push @tmpfiles, $tmpgtilst };
	$subgtierror = SubstituteGTI("tmp_clfile_nobaddet_newgti.evt","GTI",$tmpgti,$histstr,$chatter);
	die "ERROR: GTI substitution FAILED" if ($subgtierror==-1);
        if ($cleanup) { push @tmpfiles, $tmpgti };
        # Extract the total spectrum
        $cmd = "extractor ".
          "filename='tmp_clfile_nobaddet_newgti.evt' ".
          "eventsout='NONE' ". "imgfile='NONE' ".
          "phafile='$totspec' ".
          "wtmapb='no' ". "fitsbinlc='NONE' ". "binlc='16.0' ".
          "lcthresh='0.0' ". "regionfile='NONE' ". "timefile='NONE' ". 
          "xcolf='RAWX' ". "ycolf='RAWY' ". "tcol='TIME' ". "ecol='PI' ".
          "xcolh='RAWX' ". "ycolh='RAWY' ". "phamax='TLMAX' ". 
          "xfkey='TLMAX' ". "yfkey='TLMAX' ". 
          "xhkey='TLMAX' ". "yhkey='TLMAX' ". "clobber='$clobstr' ";
	print "$cmd\n" if ($chatter >= 3);
	system($cmd);
	die "ERROR: extractor failed to create total spectrum" if ($?);
	die "ERROR: new total spectrum not produced" if (! -f $totspec);
        my @weights = map {$_ / $exp_bkg} @weights;
        my @exp_weights = map {$_ / $exp_bkg} @exp_weights;
        my @supweights = map {$_ / $exp_bkg} @supweights;
	my @supexp_weights = map {$_ / $exp_bkg} @supexp_weights;
	if ($verbose) {
	    print "EXP_BKG: $exp_bkg\n";
	    for (my $jj=0; $jj<$nibg; $jj+=1) {
		print "ID: $bg_ngt_ids[$jj], IBG: $bg_ngt_ibgs[$jj], SPEC: $bg_ngt_names[$jj], WT: $weights[$jj], EXP_WT: $exp_weights[$jj]\n";
	    }
	    for (my $jj=0; $jj<$nnz; $jj+=1) {
		print "ID: $bg_day_ids[$jj], NZ: $bg_day_nzs[$jj], SPEC: $bg_day_names[$jj], $supweights[$jj]\n";
	    }
	}
        # Move copies of library spectra to local directory --
        # mathpha doesn't like full path, and doesn't like long strings
        for (my $ll=0; $ll<$nibg; $ll+=1) {
	    if ($weights[$ll] > 0.0) {
                $namein = "$bkglibdir" . "/" . "$bg_ngt_names[$ll]";
                $nameout = "tmp_ngtbg" ."$ll" .".pha";
		$cmd = "ftcopy ". "infile='$namein' ". "outfile='$nameout' ". "clobber=yes";
		print "$cmd\n" if ($chatter >= 3);
		system($cmd);
                die "ERROR: ftcopy failed to copy nighttime library spectrum" if ($?);
                die "ERROR: nighttime library spectrum $namein not copied" if (! -f $nameout );
                if ($cleanup) { push @tmpfiles, $nameout };
	    }
	}
        for (my $ll=0; $ll<$nnz; $ll+=1) {
	    if ($supweights[$ll] > 0.0) {
                $namein = "$bkglibdir" . "/" . "$bg_day_names[$ll]";
                $nameout = "tmp_daybg" ."$ll" .".pha";
		$cmd = "ftcopy ". "infile='$namein' ". "outfile='$nameout' ". "clobber=yes";
		print "$cmd\n" if ($chatter >= 3);
		system($cmd);
                die "ERROR: ftcopy failed to copy daytime library spectrum" if ($?);
                die "ERROR: nighttime library spectrum $namein not copied" if (! -f $nameout );
                if ($cleanup) { push @tmpfiles, $nameout };
	    }
	}
        # Make the weighted background spectrum
	my $minwt = $parms{minweight};
	$wterror = CreateWeightedBkgSpec(\@bg_ngt_expos,\@weights,\@exp_weights,\@bg_day_expos,\@supweights,\@supexp_weights,\@bg_ngt_names,\@bg_day_names,\$bkgspec,\$bkgspec_ngt,\$bkgspec_day,\$minwt,\$chatter,\$clobstr);
	die "ERROR: Creation of weighted Background spectra FAILED" if ($wterror==-1);	
    }
    return $status;
}

sub parse_fpm_list ($) {
  my ($fpmstr) = (@_);
  my (@fpmlist);

  my @fpmvals = split(/,/,"$fpmstr");
  foreach my $fpmval (@fpmvals) {
    if ($fpmval =~ m/^\d+$/) {
      push @fpmlist, $fpmval;
    } elsif ($fpmval =~ m/^(\d+)-(\d+)$/) {
      push @fpmlist, ($1 .. $2);
    } else {
      die "ERROR: FPM slice number specification $fpmval not recognized";
    }
  }

  return @fpmlist;
}

sub createStartStopGTI ($$$$$) {

    # Create a new GTI file from the GTI extension of an existing event file, bounded such that dtmin<STOP-START<dtmax
    my $clfile = shift;
    my $dtmin = shift;
    my $dtmax = shift;
    my $histstr = shift;
    my $chatter = shift;

    my $status = 0;
    my $hdrfile = "tmp_hdrfile";
    my $cdfile = "tmp_cdfile";
    my $datafile = "tmp_datafile";

    $ingti = SimpleFITS->open("$clfile", type=>"table", access=>"read");
    die "ERROR: Failed to open cleaned event file" if (! $ingti );
    $ingti->move("GTI");
    $ingti->readkey("NAXIS2",$inngti);
    $ingti->readkey("MJDREFI",$mjdrefi);
    $ingti->readkey("MJDREFF",$mjdreff);
    $ingti->readkey("TIMEZERO",$timezero);
    if (length($timezero) == 0) { $ingti->setstatus(0) };
    @ingtistart = $ingti->readcol("START",{type=>TDOUBLE});
    @ingtistop = $ingti->readcol("STOP",{type=>TDOUBLE});
    @indtgti = $ingti->calculate("STOP-START",{type=>TDOUBLE});
    $ingti->close();

    open HDRFILE, ">", $hdrfile;
    print HDRFILE "EXTNAME 'GTI'\n";
    print HDRFILE "HDUCLAS1 'GTI' / Contains good time intervals\n";
    print HDRFILE "MJDREFI '" .$mjdrefi. "'" ."\n";
    print HDRFILE "MJDREFF '" .$mjdreff. "'" ."\n" ;
    if (length($timezero) != 0) {print HDRFILE "TIMEZERO '$timezero' / Zero-point offset for TIME column\n" };
    close HDRFILE;
    die "ERROR: GTI header file not produced in createStartStopGTI" if (! -f $hdrfile);
    if ($cleanup) { push @tmpfiles, $hdrfile };

    open CDFILE, ">", $cdfile;
    print CDFILE "START 1D s\n";
    print CDFILE "STOP 1D s\n";
    close CDFILE;
    die "ERROR: GTI coldef file not produced in createStartStopGTI" if (! -f $cdfile);
    if ($cleanup) { push @tmpfiles, $cdfile };

    open DATAFILE, ">", $datafile;

    for (my $ii=0; $ii<$inngti; $ii+=1) {
	if ($indtgti[$ii] > $dtmax) {
	    my $nextstart = $ingtistart[$ii];
            my $nn = int($indtgti[$ii]/$dtmax);
	    for (my $jj=0; $jj<$nn; $jj+=1) {
		my $newstart = $nextstart;
                my $newstop = $nextstart + $dtmax;
                $nextstart = $newstop;
		print DATAFILE "$newstart $newstop\n";
	    }
	    if (fmod($indtgti[$ii],$dtmax) > $dtmin) {
		$newstart = $nextstart;
                $newstop = $ingtistop[$ii];
		print DATAFILE "$newstart $newstop\n";
	    }

	} elsif ($indtgti[$ii] > $dtmin) {
	    print DATAFILE "$ingtistart[$ii] $ingtistop[$ii]\n";
	}
    }

    close DATAFILE;
    die "ERROR: GTI data file not produced in createStartStopGTI" if (! -f $datafile);
    if ($cleanup) { push @tmpfiles, $datafile };

    $cmd = "ftcreate ".
      "cdfile='tmp_cdfile' ".
      "datafile='tmp_datafile' ".
      "outfile='tmp_newgti.gti' ".
      "headfile='tmp_hdrfile' ".
      "clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: ftcreate failed to make new gti file in createStartStopGTI" if ($?);
    die "ERROR: new GTI file not produced in createStartStopGTI" if (! -f "tmp_newgti.gti");
    if ($cleanup) { push @tmpfiles, "tmp_newgti.gti" };

    return $status;
}

sub SubstituteGTI ($$$$$) {

    # Subsititue and apply a new GTI to an event file
    my $infile = shift;
    my $ext_to_add = shift;
    my $gti_to_add = shift;
    my $histstr = shift;
    my $chatter = shift;

    my $status = 0;
    my $ext_to_del = "GTI";

    $cmd = "ftdelhdu infile='$infile"."[GTI]' ".
      "outfile='NONE' ".
      "confirm='yes' ".
      "clobber='yes' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: failed to delete GTI hdu in SubstituteGTI" if ($?);

    $cmd = "ftappend infile='$gti_to_add"."[$ext_to_add]' ".
      "outfile='$infile' ".
      "history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: failed to append new hdu in SubstituteGTI" if ($?);

    $cmd = "ftcopy infile='$infile"."[EVENTS][gtifilter()]' ".
      "outfile='$infile' ".
      "copyall='yes' ".
      "clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: failed to filter on new GTI in SubstituteGTI" if ($?);

    return $status;
}

sub RateGTI ($$$$$) {

    # Calculate and return an event rate from an input file filtered with an inpout GTI
    # The delta-time is pre-calculated - could use gtisum locally
    # A scale-factor (here to account for <52 MPUs) is applied
    my $infile = shift;
    my $gtifile = shift;
    my $dt = shift;
    my $scale = shift;
    my $histstr = shift;
    my $chatter = shift;
    die "ERROR: RateGTI: scale factor must be >0" if ( $scale<=0  );
    die "ERROR: RateGTI: delta-t must be >0" if ( $dt<=0  );
    
    my $gtiexpr = "gtifilter(\"$gtifile\")";
    $cmd = "ftcopy infile='$infile"."[EVENTS][$gtiexpr]' ".
      "outfile='tmp_single_gti.evt' ".
      "copyall='yes' ".
      "clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: failed to filter on single GTI in RateGTI" if ($?);
    die "ERROR: new single gti event file not produced in RateGTI" if (! -f "tmp_single_gti.evt");
    if ($cleanup) { push @tmpfiles, "tmp_single_gti.evt" };

    $tmp_single_gti_evt = SimpleFITS->open("tmp_single_gti.evt", type=>"table", access=>"read");
    $tmp_single_gti_evt->move("EVENTS");
    $tmp_single_gti_evt->readkey("NAXIS2",$num_single_gti_evt);
    $tmp_single_gti_evt->close();

    my $gtirate = $num_single_gti_evt / $scale / $dt;

    return $gtirate;
}

sub NewFindIBGID {

    # Fine the 3C50_RGv5 night spectrum ID from ibg (rate1) and hrej (rate2) 
    my ($bg_ngt_ids_ref, $ibglo_ref, $ibghi_ref, $hrejlo_ref, $hrejhi_ref, $nibg_ref, $ibgmax_ref, $hrejmax_ref, $rate1_ref, $rate2_ref) = @_;
    my @bg_ngt_ids = @{ $bg_ngt_ids_ref };
    my @ibglo = @{ $ibglo_ref };
    my @ibghi = @{ $ibghi_ref };
    my @hrejlo = @{ $hrejlo_ref };
    my @hrejhi = @{ $hrejhi_ref };
    my $nibg = ${ $nibg_ref };
    my $ibgmax = ${ $ibgmax_ref };
    my $hrejmax = ${ $hrejmax_ref };
    my $rate1 = ${ $rate1_ref };
    my $rate2 = ${ $rate2_ref };

    my $id12 = "00";
    my $ii = 0;
    my $flag = 0;
    
    if (($rate1<=$ibgmax) && ($rate2<=$hrejmax)) {
	$ii = 0;
	$flag = 0;
	while (($flag==0) && ($ii<$nibg)) {
	    if (($rate1>$ibglo[$ii]) && ($rate1<=$ibghi[$ii])) {
		if (($rate2>=$hrejlo[$ii]) && ($rate2<$hrejhi[$ii])) {
		    $id12 = $bg_ngt_ids[$ii];
		    $flag = 1
		}
	    }
	    $ii+=1
	}
    } else {
	$id12 = "00"
    }      			
    return $id12;
}

sub NewFindNZID {

    # Get the 3C50_RGv5 day spectrum ID from nz rate
    my ($bg_day_ids_ref, $nzlo_ref, $nzhi_ref, $nnz_ref, $nzmax_ref, $rate_ref) = @_;
    my @bg_day_ids = @{ $bg_day_ids_ref };
    my @nzlo = @{ $nzlo_ref };
    my @nzhi = @{ $nzhi_ref };
    my $nnz = ${ $nnz_ref };
    my $nzmax = ${ $nzmax_ref };
    my $rate = ${ $rate_ref };

    my $id33 = "00";
    my $ii = 0;
    my $flag = 0;
	
    if ($rate>$nzmax) {
	$id33 = $bg_day_ids[$nnz-1];
    } else {
	$ii = 0;
	$flag = 0;
	while (($flag==0) && ($ii<$nnz-1)) {
	    if (($rate>$nzlo[$ii]) && ($rate<=$nzhi[$ii])) {
		$id33 = $bg_day_ids[$ii];
		$flag = 1;
	    }
	    $ii+=1;
	}
    }      			
    return $id33;
}

sub CreateWeightedBkgSpec {

    # Use mathpha to create a weighted spectrum with weighted exposure
    # Weights and library spectrum names are input
    my ($exposngt_ref, $wts_ref, $expwts_ref, $exposday_ref, $supwts_ref, $supexpwts_ref, $ngtbg_names_ref, $daybg_names_ref, $bkgspec_ref, $bkgspecngt_ref, $bkgspecday_ref, , $minwt_ref, $chat_ref, $clob_ref) = @_;
    my @exposngt = @{ $exposngt_ref }; 
    my @wts = @{ $wts_ref };
    my @expwts = @{ $expwts_ref };
    my @exposday = @{ $exposday_ref }; 
    my @supwts = @{ $supwts_ref };
    my @supexpwts = @{ $supexpwts_ref };
    my @ngtbg_names = @{ $ngtbg_names_ref };
    my @daybg_names = @{ $daybg_names_ref };
    my $bkgspec = $ { $bkgspec_ref };
    my $bkgspec_ngt = $ { $bkgspecngt_ref };
    my $bkgspec_day = $ { $bkgspecday_ref };
    my $minwt = $ { $minwt_ref };
    my $chatter = $ { $chat_ref };
    my $clobber = $ { $clob_ref };

    my $bgspec_expo = 0.0;
    for (my $ii=0; $ii<scalar @exposngt; $ii+=1) {
	$bgspec_expo += $expwts[$ii] * $exposngt[$ii];
    }

    my $bgspec_supexpo = 0.0;
    for (my $ii=0; $ii<scalar @exposday; $ii+=1) {
	$bgspec_supexpo += $supexpwts[$ii] * $exposday[$ii];
    }

    # Hierarchical queueing system for building mathpha not-too-big expressions 
    # maximum number of spectra per mathpha call
    my $nmax = 8;
    # initialize index within each mathpha call
    my $numspec = 0;
    my $numspec_d = 0;
    # initialize index of each mathpha call
    my $numsplit = 0;
    my $numsplit_d = 0;
    my $mathpha_expr = "";
    my $mathpha_expr_d = "";
    my $mathpha_expr_q = "";
    my $mathpha_expr_qd = "";
    my $mathpha_expr_com = "";
    my $ilast = -1;
    for (my $ii=0; $ii<scalar @wts; $ii+=1) {
	if ($wts[$ii] > $minwt) {
	    $ilast = $ii;
	}
    }
    die "ERROR: no non-zero night backlib weights found in CreateWeightedBkgSpec" if ($ilast==-1);
    my $ilast_d = -1;
    for (my $ii=0; $ii<scalar @supwts; $ii+=1) {
	if ($supwts[$ii] > $minwt) {
	    $ilast_d = $ii;
	}
    }

    # loop over the night library
    for (my $ii=0; $ii<=$ilast; $ii+=1) {
	if ($wts[$ii] > $minwt) {
	    $wt_str = mathpha_str($wts[$ii]);
	    if ($numspec==0) {
		# start new queue
		$numsplit += 1;
		$tempspecname =  "tmpbkg"."$numsplit".".pi";
		if ($cleanup) { push @tmpfiles, $tempspecname };
                $mathpha_expr_q = "";		    
		$expr_to_add = $wt_str."*tmp_ngtbg"."$ii".".pha";
		$numspec += 1;
	    } else {
		# add to existing queue
		$expr_to_add = "+".$wt_str."*tmp_ngtbg"."$ii".".pha";
	    }
	    $mathpha_expr_q = $mathpha_expr_q.$expr_to_add;
	    if ( ($numspec < $nmax-1) && ($ii < $ilast) ) {
		# continue
		$numspec+=1;
	    } else {
		# close the queue, make the sub-spectrum, add to master expr  	
		$cmd = "mathpha ".
		    "expr='$mathpha_expr_q' ". "units='R' ".
		    "outfil='$tempspecname' ". "exposure='$bgspec_expo' ".
		    "areascal='%' ". "properr='yes' ". "errmeth='gaussian' ".
		    "ncomments='0' ". "clobber=yes";
		system($cmd);
		die "ERROR: mathpha failed to create temp night bkg sub-spectrum" if ($?);
		if ($numsplit==1) {
		    $expr_to_add = $tempspecname;
		} else { 
		    $expr_to_add = "+".$tempspecname;
		}
		$mathpha_expr = $mathpha_expr.$expr_to_add;
		$numspec = 0;
	    } #endif ( ($numspec < $nmax-1) && ($ii < $ilast) )
	} #endif ($wts[$ii] > minwt)
    } #end loop over the night library
    
    # final night spectrum
    $cmd = "mathpha ".
	"expr='$mathpha_expr' ". "units='R' ".
	"outfil='tmp_bkgspec_ngt.pi' ". "exposure='$bgspec_expo' ".
	"areascal='%' ". "properr='yes' ". "errmeth='gaussian' ".
	"ncomments='0' ". "clobber='yes'";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: mathpha failed to create night bkg spectrum" if ($?);
    die "ERROR: night bkg spectrum not produced" if (! -f "tmp_bkgspec_ngt.pi");
    if ($cleanup) { push @tmpfiles, "tmp_bkgspec_ngt.pi"};
    $cmd = "ftcopy infile='tmp_bkgspec_ngt.pi' ".
      "outfile='$bkgspec_ngt'  ".
      "copyall='yes' ".
      "clobber='$clobber' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: ftcopy failed to copy temp file to night bkg spectrum" if ($?);
    die "ERROR: night bkg spectrum not renamed" if (! -f "$bkgspec_ngt");
    
    if ($ilast_d > -1) {
	# loop over the day library
	for (my $ii=0; $ii<=$ilast_d; $ii+=1) {
	    if ($supwts[$ii] > $minwt) {
		$supwt_str = mathpha_str($supwts[$ii]);
		if ($numspec_d==0) {
		    # start new queue
		    $numsplit_d += 1;
		    $tempspecname =  "tmpsupbkg"."$numsplit".".pi";
		    if ($cleanup) { push @tmpfiles, $tempspecname };
		    $mathpha_expr_qd = "";		    
		    $expr_to_add = $supwt_str."*tmp_daybg"."$ii".".pha";
		    $numspec_d += 1;
		} else {
		    # add to queue
		    $expr_to_add = "+".$supwt_str."*tmp_daybg"."$ii".".pha";
		}
                $mathpha_expr_qd = $mathpha_expr_qd.$expr_to_add;
		if ( ($numspec_d < $nmax-1) && ($ii < $ilast_d) ) {
		    # continue
		    $numspec_d+=1;
		} else {
		    # close the queue, make the sub-spec, add to master expr	
		    $cmd = "mathpha ".
			"expr='$mathpha_expr_qd' ". "units='R' ".
			"outfil='$tempspecname' ". "exposure='$bgspec_supexpo' ".
			"areascal='%' ". "properr='yes' ". "errmeth='gaussian' ".
			"ncomments='0' ". "clobber=yes";
		    print "$cmd\n" if ($chatter >= 3);
		    system($cmd);
		    die "ERROR: mathpha failed to create temp day bkg sub-spectrum" if ($?);
		    if ($numsplit_d==1) {
			$expr_to_add = $tempspecname;
		    } else { 
			$expr_to_add = "+".$tempspecname;
		    }
		    $mathpha_expr_d = $mathpha_expr_d.$expr_to_add;
		    $numspec_d = 0;
		} #endif ( ($numspec_d < $nmax-1) && ($ii < $ilast_d) )
	    } #endif ($supwts[$ii] > minwt)
	} #end loop over the day library
		
	# final day spectrum
	$cmd = "mathpha ".
	    "expr='$mathpha_expr_d' ". "units='R' ".
	    "outfil='tmp_bkgspec_day.pi' ". "exposure='$bgspec_supexpo' ".
	    "areascal='%' ". "properr='yes' ". "errmeth='gaussian' ".
	    "ncomments='0' ". "clobber='yes'";	
	print "$cmd\n" if ($chatter >= 3);
	system($cmd);	
	die "ERROR: mathpha failed to create day bkg spectrum" if ($?);
	die "ERROR:  day bkg spectrum not produced" if (! -f "tmp_bkgspec_day.pi");
	if ($cleanup) { push @tmpfiles, "tmp_bkgspec_day.pi"};
	$cmd = "ftcopy infile='tmp_bkgspec_day.pi' ".
	  "outfile='$bkgspec_day'  ".
	  "copyall='yes' ".
	  "clobber='$clobber' "."history='$histstr' "."chatter='$chatter' ";
	print "$cmd\n" if ($chatter >= 3);
	system($cmd);
	die "ERROR: ftcopy failed to copy temp file day bkg spectrum" if ($?);
	die "ERROR: day bkg spectrum not renamed" if (! -f "$bkgspec_day");
	
	# final composite spectrum
	$mathpha_expr_com = "tmp_bkgspec_ngt.pi"."+"."tmp_bkgspec_day.pi";
	$cmd = "mathpha ".
	    "expr='$mathpha_expr_com' ". "units='R' ".
	    "outfil='tmp_bkg.pi' ". "exposure='$bgspec_expo' ".
	    "areascal='%' ". "properr='yes' ". "errmeth='gaussian' ".
	    "ncomments='0' ". "clobber=yes";	
	print "$cmd\n" if ($chatter >= 3);
	system($cmd);	
	die "ERROR: mathpha failed to create composite bkg spectrum" if ($?);
	die "ERROR:  Composite bkg spectrum not produced" if (! -f "tmp_bkg.pi");	
    } else {
	$cmd = "ftcopy ".
	    "infile='tmp_bkgspec_ngt.pi' ".
	    "outfile='tmp_bkg.pi' ".
	    "copyall='yes' ".
	    "clobber='yes' "."history='$histstr' "."chatter='$chatter' ";
	print "$cmd\n" if ($chatter >= 3);
	system($cmd);    
        die "ERROR: ftcopy failed to copy temp to bkg spectrum" if ($?);
    } #endif ($ilast_d > -1) {
    if ($cleanup) { push @tmpfiles, "tmp_bkg.pi" };
    $cmd = "ftcopy infile='tmp_bkg.pi' ".
      "outfile='$bkgspec'  ".
      "copyall='yes' ".
      "clobber='$clobber' "."history='$histstr' "."chatter='$chatter' ";
    print "$cmd\n" if ($chatter >= 3);
    system($cmd);
    die "ERROR: ftcopy failed to copy temp file to total bkg spectrum" if ($?);
    die "ERROR: total bkg spectrum not renamed" if (! -f "$bkgspec");
    
    die "ERROR: background spectrum not produced in CreateWeightedBkgSpec" if (! -f $bkgspec);
    
    return $status;
}

sub mathpha_str($) {

    # Convert weights from exponential to fractional notation for use in mathpha
    my $mstr = shift;
    my $idx = index($mstr, "e");
    if ($idx != -1) {
	$nzeros = (substr $mstr, -2)-1;
        $nnum1 = substr $mstr, 0, 1;
        $nnum2 = substr $mstr, 2, $idx-2;
	my $zeros = "0.";
	for (my $ii=0; $ii<$nzeros; $ii+=1) {$zeros=$zeros."0"}
	$newnum = $zeros.$nnum1.$nnum2;
	return $newnum;
    } else {
	return $mstr;
    }
}

sub CheckGTIName ($) {

    # If first GTI/STDGTI extension is STDGTI rename to GTI
    my $status = 0;
    my $infile = shift;
    $ingti = SimpleFITS->open("$infile", type=>"table", access=>"read");
    $ingti->nhdu($n_hdu);
    $ingti->move(1);
    for (my $ihdu=0; $ihdu<$n_hdu-1; $ihdu+=1) {
	$ingti->move("+1");
	$ingti->readkey("EXTNAME",$this_extname);
	if ($this_extname eq "GTI")  {
	    last;
	} elsif ($this_extname eq "STDGTI") {
	    $cmd = "fthedit infile='$infile"."[STDGTI]' ".
		"keyword='EXTNAME' ".
		"operation='add' ".
		"value='GTI' ".
		"history='$histstr' "."chatter='$chatter' ";
	    print "$cmd\n" if ($chatter >= 3);
	    system($cmd);
	    $status = 1;
	    last;
	}
    }

    if (not (($this_extname eq "GTI") | ($this_extname eq "STDGTI")) ) {
	$status = -1;
    }
    return $status;
}
